{-# LANGUAGE NamedFieldPuns, OverloadedStrings #-}
module Join
  ( Config
  , config
  , run ) where

import Control.Applicative ((<|>), optional)
import Data.Attoparsec.Text (char, parseOnly, sepBy, takeWhile1)
import Data.Char (toLower)
import Data.Text as Text (Text, pack)
import Data.Text.Encoding (decodeUtf8)
import Data.Vector as Vector (toList)
import GEODE.Metadata
  ( ArticleRecord, ColumnsMapping, JoinContext(..), JoinMethod, ReadTSV(..)
  , Renamers, WriteTSV(..), for, getHeader, leftJoin, rightJoin )
import Options.Applicative
  ( Parser, long, metavar, option, short, str, strArgument, strOption, value )
import System.Exit (die)
import Text.Printf (printf)

data Config = Config
  { leftTSV :: FilePath
  , rightTSV :: FilePath
  , outputTSV :: Maybe FilePath
  , on :: Text
  , how :: Maybe String
  , renamePatterns :: Maybe String }

config :: Parser Config
config = Config
  <$> strArgument (metavar "LEFT_TSV")
  <*> strArgument (metavar "RIGHT_TSV")
  <*> option (optional str) (long "output" <> short 'o' <> value Nothing)
  <*> strOption (long "on" <> short 'O')
  <*> (option (optional str) (long "how" <> short 'H' <> value Nothing))
  <*> (option (optional str) (long "rename" <> short 'r' <> value Nothing))

dup :: Applicative f => a -> f (a, a)
dup a = pure (a, a)

parseColumns :: Text -> IO ColumnsMapping
parseColumns "ARTICLES" = pure (articleHeader, articleHeader)
  where
    articleHeader = decodeUtf8 <$>
      Vector.toList (getHeader (for :: ArticleRecord))
parseColumns s = either die (pure . unzip) $ parseOnly columns s
  where
    columns = (mapping <|> (columnName >>= dup)) `sepBy` char ','
    mapping = (,) <$> columnName <* char '=' <*> columnName
    columnName = takeWhile1 (not . (`elem` ['=', ',']))

parseRenamers :: Maybe String -> IO Renamers
parseRenamers Nothing = dup id
parseRenamers (Just s) =
  case break (== ',') s of
    (left, ',':right) -> pure (Text.pack . printf left, Text.pack . printf right)
    _ -> die "--rename option expects to comma-separated printf pattern (with only one '%s' in each)"

parseMethod :: Maybe String -> IO JoinMethod
parseMethod Nothing = pure leftJoin
parseMethod (Just s) = readMethod $ toLower <$> s
  where
    readMethod "left" = pure leftJoin
    readMethod "right" = pure rightJoin
    readMethod e = die ("Unknown join method '" ++ e ++ "'")

run :: Config -> IO ()
run (Config {leftTSV, rightTSV, outputTSV, how, on, renamePatterns}) = do
  method <- parseMethod how
  context <- JoinContext <$> parseColumns on <*> parseRenamers renamePatterns
  (method context <$> (readTSV leftTSV) <*> (readTSV rightTSV))
    >>= (maybe (writeTSV ()) writeTSV outputTSV)
